package org.processmining.behavioralpatternmining.algorithms;


import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.RecursiveAction;

import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.impl.XLogImpl;
import org.processmining.behavioralpatternmining.models.BehavioralPattern;
import org.processmining.behavioralpatternmining.models.CombinationType;
import org.processmining.behavioralpatternmining.models.Leaf;
import org.processmining.behavioralpatternmining.models.Node;
import org.processmining.behavioralpatternmining.models.Operator;
import org.processmining.behavioralpatternmining.models.ProcessTreeSet;
import org.processmining.behavioralpatternmining.parameters.BehavioralPatternParameters;
import org.processmining.behavioralpatternmining.util.ProcessTreeUtils;
import org.processmining.framework.plugin.PluginContext;


public class BPRecursiveAction extends RecursiveAction {

	private static final long serialVersionUID = 1612357811227379936L;

	private ProcessTreeSet lpmSet;
	private BehavioralPattern lpm;
	private BehavioralPattern lpmOther;
	private XLog projection;
	private Map<Integer, Integer> projectionMapping;
	private Set<Integer> projectionMap;
	private CombinationType type;
	private Map<String, ProcessTreeSet> gammASetsMap;
	private Map<String, ProcessTreeSet> gammaSetsMap;
	private Map<String, BehavioralPattern> evaluatedLPMs;
	private Set<BehavioralPattern> gammA;
	private ProcessTreeEvaluator evaluator;
	private BehavioralPatternParameters params;
	private PluginContext context;

	public BPRecursiveAction(ProcessTreeSet lpmSet, BehavioralPattern lpm, BehavioralPattern lpmOther,
			XLog projection, Map<Integer, Integer> projectionMapping, Set<Integer> projectionMap, CombinationType type,
			Map<String, ProcessTreeSet> gammASetsMap, Map<String, ProcessTreeSet> gammaSetsMap,
			Map<String, BehavioralPattern> evaluatedLPMs, Set<BehavioralPattern> gammA,
			ProcessTreeEvaluator evaluator, BehavioralPatternParameters params, PluginContext context) {
		this.lpmSet = lpmSet;
		this.lpm = lpm;
		this.lpmOther = lpmOther;
		this.projection = projection;
		this.projectionMapping = projectionMapping;
		this.projectionMap = projectionMap;
		this.type = type;
		this.gammaSetsMap = gammaSetsMap;
		this.gammASetsMap = gammASetsMap;
		this.evaluatedLPMs = evaluatedLPMs;
		this.gammA = gammA;
		this.evaluator = evaluator;
		this.params = params;
		this.context = context;
	}

	protected void compute() {
		if (type == CombinationType.SEQ || type == CombinationType.AND || type == CombinationType.LOOP) {

			List<Object> linearFormatOther = lpmOther.getLinearFormat();
			List<Object> linearFormatNew = ProcessTreeUtils.copyLinearFormat(lpm.getRoot());
			Leaf leaf = (Leaf) linearFormatNew.get(lpmSet.getPositionCombi());
			leaf.setDepth(leaf.getDepth() + 1);
			Leaf leafOther = new Leaf(((Leaf) linearFormatOther.get(lpmSet.getPositionCombi())).getName());
			leafOther.setDepth(leaf.getDepth());
			leafOther.setRepresLeaf(leafOther.getName());

			linearFormatNew.add(lpmSet.getPositionCombi() + 1, leafOther);
			Operator operator;
			if (type == CombinationType.SEQ)
				operator = new Operator("Seq");
			else if (type == CombinationType.AND)
				operator = new Operator("And");
			else
				operator = new Operator("Loop");
			operator.setChild1(leaf);
			operator.setChild2(leafOther);
			if (leaf.getParent() != null) {
				operator.setParent(leaf.getParent());
				if (leaf.getParent().getChild1() == leaf)
					leaf.getParent().setChild1(operator);
				else
					leaf.getParent().setChild2(operator);
			}
			linearFormatNew.add(lpmSet.getPositionCombi() + 2, operator);

			operator.setRepresLeaf(leaf.getName() < leafOther.getName() ? leaf.getName() : leafOther.getName());
			Operator parent = operator.getParent();
			while (parent != null) {
				parent.setRepresLeaf(parent.getChild1().getRepresLeaf() < parent.getChild2().getRepresLeaf()
						? parent.getChild1().getRepresLeaf()
						: parent.getChild2().getRepresLeaf());
				parent = parent.getParent();
			}
			String represStringLPM = ProcessTreeUtils
					.represString((Node) linearFormatNew.get(linearFormatNew.size() - 1));
			BehavioralPattern lpmNew;
			if (evaluatedLPMs.containsKey(represStringLPM)) {
				BehavioralPattern lpmExist = evaluatedLPMs.get(represStringLPM);
				lpmNew = new BehavioralPattern();
				lpmNew.setFrequency(lpmExist.getFrequency(), params.getDiscoveryLog().size());
				lpmNew.setLanguageRatio(lpmExist.getLanguageRatio());
				lpmNew.setLanguage(lpmExist.getLanguage());
				lpmNew.setLanguageSeen(lpmExist.getLanguageSeen());
				lpmNew.setProjectionMap(lpmExist.getProjectionMap());

			} else {
				lpmNew = evaluator.evaluateProcessTreeOnLog(context,
						(Node) linearFormatNew.get(linearFormatNew.size() - 1), projection, projectionMapping, false);
				evaluatedLPMs.put(represStringLPM, lpmNew);
			}
			int leafDepth = leaf.getDepth();
			lpmNew.setRepresString(represStringLPM);
			lpmNew.setLinearFormat(linearFormatNew);
			lpmNew.setDepth(lpm.getDepth() > leafDepth ? lpm.getDepth() : leafDepth);
			lpmNew.setSeed1(lpm);
			lpmNew.setSeed2(lpmOther);
			lpmNew.setDefiningOperator(type);
			lpmNew.setRoot((Node) linearFormatNew.get(linearFormatNew.size() - 1));
			//System.out.println(lpmNew);
			if (lpmNew.getSupport() >= params.getSupportMinimum()) {
				if (lpmNew.getLanguageRatio() >= params.getLanguageFitMinimum()) {
					gammA.add(lpmNew);
					removeConstructionTree(lpmNew);
				}

				int currentLeftDepth = 0;

				for (int j = 0; j < linearFormatNew.size(); j++)
					if (linearFormatNew.get(j) instanceof Leaf) {
						Leaf task = (Leaf) linearFormatNew.get(j);

						int taskDepth = task.getDepth();
						if (taskDepth >= currentLeftDepth && taskDepth >= lpm.getDepth() - 1) {
							currentLeftDepth = taskDepth;
							String represStringSet = represStringGen(linearFormatNew, j);
							List<BPRecursiveAction> subtasks = null;
							synchronized (represStringSet) {
								if (gammASetsMap.containsKey(represStringSet)) {
									ProcessTreeSet set = gammASetsMap.get(represStringSet);
									if (set.isExtensible()) {

										subtasks = createFreqSubtasks(set, lpmNew);
									} 
									set.add(lpmNew);
								} else {
									ProcessTreeSet newSet = new ProcessTreeSet();
									gammASetsMap.put(represStringSet, newSet);
									newSet.setDepth(lpmNew.getDepth());
									newSet.setPositionCombi(j);
									newSet.setRepresString(represStringSet);
									newSet.setExtensible(taskDepth + 1 <= params.getDepthMax());
									newSet.setContainsFreq(true);
									newSet.add(lpmNew);

								}
							}
							if (subtasks != null)
								invokeAll(subtasks);

						}

					}

			} else {
				int currentLeftDepth = 0;

				for (int j = 0; j < linearFormatNew.size(); j++)
					if (linearFormatNew.get(j) instanceof Leaf) {
						Leaf task = (Leaf) linearFormatNew.get(j);

						int taskDepth = task.getDepth();
						if (taskDepth >= currentLeftDepth && taskDepth >= lpm.getDepth() - 1) {
							currentLeftDepth = taskDepth;
							String represStringSet = represStringGen(linearFormatNew, j);
							List<BPRecursiveAction> subtasks = null;
							synchronized (represStringSet) {
								if (gammaSetsMap.containsKey(represStringSet)) {
									ProcessTreeSet set = gammaSetsMap.get(represStringSet);
									if (set.isExtensible()) {
										subtasks = createInfreqSubtasks(set, lpmNew);
									}
									set.add(lpmNew);
								} else {
									ProcessTreeSet newSet = new ProcessTreeSet();
									newSet.setDepth(lpmNew.getDepth());
									newSet.setPositionCombi(j);
									newSet.setRepresString(represStringSet);
									newSet.setExtensible(taskDepth + 1 <= params.getDepthMax());
									newSet.setContainsFreq(false);
									gammaSetsMap.put(represStringSet, newSet);
									newSet.add(lpmNew);
								}
							}
							if (subtasks != null)
								invokeAll(subtasks);
						}

					}
			}

		} else if (type == CombinationType.XOR) {
			List<Object> linearFormatOther = lpmOther.getLinearFormat();
			List<Object> linearFormatNew = ProcessTreeUtils.copyLinearFormat(lpm.getRoot());
			Leaf leaf = (Leaf) linearFormatNew.get(lpmSet.getPositionCombi());
			leaf.setDepth(leaf.getDepth() + 1);
			Leaf leafOther = new Leaf(((Leaf) linearFormatOther.get(lpmSet.getPositionCombi())).getName());
			leafOther.setDepth(leaf.getDepth());
			leafOther.setRepresLeaf(leafOther.getName());

			linearFormatNew.add(lpmSet.getPositionCombi() + 1, leafOther);
			Operator operator;

			operator = new Operator("Xor");

			operator.setChild1(leaf);
			operator.setChild2(leafOther);
			if (leaf.getParent() != null) {
				if (leaf.getParent().getChild1() == leaf)
					leaf.getParent().setChild1(operator);
				else
					leaf.getParent().setChild2(operator);
			}
			linearFormatNew.add(lpmSet.getPositionCombi() + 2, operator);

			operator.setRepresLeaf(leaf.getName() < leafOther.getName() ? leaf.getName() : leafOther.getName());
			Operator parent = operator.getParent();
			while (parent != null) {
				parent.setRepresLeaf(parent.getChild1().getRepresLeaf() < parent.getChild2().getRepresLeaf()
						? parent.getChild1().getRepresLeaf()
						: parent.getChild2().getRepresLeaf());
				parent = parent.getParent();
			}

			Set<List<Integer>> language = new HashSet<List<Integer>>(lpm.getLanguage());
			language.addAll(lpmOther.getLanguage());

			Set<List<Integer>> languageSeen = new HashSet<List<Integer>>(lpm.getLanguageSeen());
			languageSeen.addAll(lpmOther.getLanguageSeen());

			BehavioralPattern lpmNew = new BehavioralPattern();
			lpmNew.setFrequency(projectionMap.size(), params.getDiscoveryLog().size());
			lpmNew.setLanguageRatio(languageSeen.size() * 1.0 / language.size());
			lpmNew.setLanguage(language);
			lpmNew.setLanguageSeen(languageSeen);
			lpmNew.setProjectionMap(projectionMap);
			lpmNew.setLinearFormat(linearFormatNew);
			lpmNew.setRoot((Node) linearFormatNew.get(linearFormatNew.size() - 1));
			int leafDepth = leaf.getDepth();
			lpmNew.setDepth(lpm.getDepth() > leafDepth ? lpm.getDepth() : leafDepth);

			String represStringLPM = ProcessTreeUtils.represString(lpmNew.getRoot());
			if (!evaluatedLPMs.containsKey(represStringLPM))
				evaluatedLPMs.put(represStringLPM, lpmNew);
			lpmNew.setRepresString(represStringLPM);
			lpmNew.setSeed1(lpm);
			lpmNew.setSeed2(lpmOther);
			lpmNew.setDefiningOperator(type);
			//System.out.println(lpmNew.toString());
			if (lpmNew.getSupport() >= params.getSupportMinimum()) {
				if (lpmNew.getLanguageRatio() >= params.getLanguageFitMinimum()) {
					gammA.add(lpmNew);
				}

				int currentLeftDepth = 0;

				for (int j = 0; j < linearFormatNew.size(); j++)
					if (linearFormatNew.get(j) instanceof Leaf) {
						Leaf task = (Leaf) linearFormatNew.get(j);

						int taskDepth = task.getDepth();
						if (taskDepth >= currentLeftDepth && taskDepth >= lpm.getDepth() - 1) {
							currentLeftDepth = taskDepth;
							String represStringSet = represStringGen(linearFormatNew, j);
							List<BPRecursiveAction> subtasks = null;
							synchronized (represStringSet) {
								if (gammASetsMap.containsKey(represStringSet)) {
									ProcessTreeSet set = gammASetsMap.get(represStringSet);
									if (set.isExtensible()) {
										subtasks = createFreqSubtasks(set, lpmNew);
									} 
										set.add(lpmNew);
								} else {
									ProcessTreeSet newSet = new ProcessTreeSet();
									gammASetsMap.put(represStringSet, newSet);
									newSet.setDepth(lpmNew.getDepth());
									newSet.setPositionCombi(j);
									newSet.setRepresString(represStringSet);
									newSet.setExtensible(taskDepth + 1 <= params.getDepthMax());
									newSet.setContainsFreq(true);	
									newSet.add(lpmNew);

								}
							}
							if (subtasks != null)
								invokeAll(subtasks);

						}

					}

			} else {
				int currentLeftDepth = 0;

				for (int j = 0; j < linearFormatNew.size(); j++)
					if (linearFormatNew.get(j) instanceof Leaf) {
						Leaf task = (Leaf) linearFormatNew.get(j);

						int taskDepth = task.getDepth();
						if (taskDepth >= currentLeftDepth && taskDepth >= lpm.getDepth() - 1) {
							currentLeftDepth = taskDepth;
							String represStringSet = represStringGen(linearFormatNew, j);
							List<BPRecursiveAction> subtasks = null;
							synchronized (represStringSet) {
								if (gammaSetsMap.containsKey(represStringSet)) {
									ProcessTreeSet set = gammaSetsMap.get(represStringSet);
									if (set.isExtensible()) {
										subtasks = createInfreqSubtasks(set, lpmNew);
									}
									set.add(lpmNew);
								} else {
									ProcessTreeSet newSet = new ProcessTreeSet();
									newSet.setDepth(lpmNew.getDepth());
									newSet.setPositionCombi(j);
									newSet.setRepresString(represStringSet);
									newSet.setExtensible(taskDepth + 1 <= params.getDepthMax());
									newSet.setContainsFreq(false);
									gammaSetsMap.put(represStringSet, newSet);
									newSet.add(lpmNew);
								}
							}
							if(subtasks != null) invokeAll(subtasks);

						}

					}
			}

		} else if (type == CombinationType.ROOT) {
//			System.out.println("Freq Set of representative string : " + lpmSet.getRepresString());
//			System.out.println("Added LPM : " + lpm.getRepresString());

			List<BPRecursiveAction> subtasks = new ArrayList<BPRecursiveAction>();
			for (int i = 0; i < lpmSet.size(); i++) {
				BehavioralPattern lpmOther = lpmSet.get(i);
				Set<Integer> intersection = new HashSet<Integer>(lpm.getProjectionMap());
				intersection.retainAll(lpmOther.getProjectionMap());

				if (intersection.size() * 1.0 / params.getDiscoveryLog().size() >= params.getSupportMinimum()
						|| lpmSet.getPositionCombi() == 0) {

					XLog projection = new XLogImpl(params.getDiscoveryLog().getAttributes());
					Map<Integer, Integer> projectionMapping = new HashMap<Integer, Integer>();
					Iterator<Integer> it = intersection.iterator();
					int j = 0;
					while (it.hasNext()) {
						int traceIndex = it.next();
						projection.add(params.getDiscoveryLog().get(traceIndex));
						projectionMapping.put(j, traceIndex);
						j++;
					}
					subtasks.add(new BPRecursiveAction(lpmSet, lpm, lpmOther, projection, projectionMapping, null,
							CombinationType.AND, gammaSetsMap, gammASetsMap, evaluatedLPMs, gammA, evaluator, params,
							context));
					subtasks.add(new BPRecursiveAction(lpmSet, lpm, lpmOther, projection, projectionMapping, null,
							CombinationType.SEQ, gammaSetsMap, gammASetsMap, evaluatedLPMs, gammA, evaluator, params,
							context));
					subtasks.add(new BPRecursiveAction(lpmSet, lpmOther, lpm, projection, projectionMapping, null,
							CombinationType.SEQ, gammaSetsMap, gammASetsMap, evaluatedLPMs, gammA, evaluator, params,
							context));
					subtasks.add(new BPRecursiveAction(lpmSet, lpm, lpmOther, projection, projectionMapping, null,
							CombinationType.LOOP, gammaSetsMap, gammASetsMap, evaluatedLPMs, gammA, evaluator, params,
							context));
					subtasks.add(new BPRecursiveAction(lpmSet, lpmOther, lpm, projection, projectionMapping, null,
							CombinationType.LOOP, gammaSetsMap, gammASetsMap, evaluatedLPMs, gammA, evaluator, params,
							context));
					subtasks.add(new BPRecursiveAction(lpmSet, lpmOther, lpm, projection, projectionMapping, null,
							CombinationType.AND, gammaSetsMap, gammASetsMap, evaluatedLPMs, gammA, evaluator, params,
							context));
				}
			}
//			XLog projectionLoopSelf = new XLogImpl(params.getDiscoveryLog().getAttributes());
//			Map<Integer, Integer> projectionMappingLoopSelf = new HashMap<Integer, Integer>();
//			Iterator<Integer> it = lpm.getProjectionMap().iterator();
//			int j = 0;
//			while (it.hasNext()) {
//				int traceIndex = it.next();
//				projectionLoopSelf.add(params.getDiscoveryLog().get(traceIndex));
//				projectionMappingLoopSelf.put(j, traceIndex);
//				j++;
//			}
//			subtasks.add(new LPMRecursiveAction(lpmSet, lpm, lpm, projectionLoopSelf, projectionMappingLoopSelf, null,
//					CombinationType.LOOP, gammaSetsMap, gammASetsMap, evaluatedLPMs, gammA, evaluator, params,
//					context));
			invokeAll(subtasks);
		}

	}

	private List<BPRecursiveAction> createFreqSubtasks(ProcessTreeSet lpmSet, BehavioralPattern lpm) {
//		System.out.println("Freq Set of representative string : " + lpmSet.getRepresString());
//		System.out.println("Added LPM : " + lpm.getRepresString());

		List<BPRecursiveAction> subtasks = new ArrayList<BPRecursiveAction>();
		for (int i = 0; i < lpmSet.size(); i++) {
			BehavioralPattern lpmOther = lpmSet.get(i);
			Set<Integer> intersection = new HashSet<Integer>(lpm.getProjectionMap());
			intersection.retainAll(lpmOther.getProjectionMap());

			if (intersection.size() * 1.0 / params.getDiscoveryLog().size() >= params.getSupportMinimum()
					|| lpmSet.getPositionCombi() == 0) {

				XLog projection = new XLogImpl(params.getDiscoveryLog().getAttributes());
				Map<Integer, Integer> projectionMapping = new HashMap<Integer, Integer>();
				Iterator<Integer> it = intersection.iterator();
				int j = 0;
				while (it.hasNext()) {
					int traceIndex = it.next();
					projection.add(params.getDiscoveryLog().get(traceIndex));
					projectionMapping.put(j, traceIndex);
					j++;
				}
				subtasks.add(new BPRecursiveAction(lpmSet, lpm, lpmOther, projection, projectionMapping, null,
						CombinationType.AND, gammaSetsMap, gammASetsMap, evaluatedLPMs, gammA, evaluator, params,
						context));
				subtasks.add(new BPRecursiveAction(lpmSet, lpm, lpmOther, projection, projectionMapping, null,
						CombinationType.SEQ, gammaSetsMap, gammASetsMap, evaluatedLPMs, gammA, evaluator, params,
						context));
				subtasks.add(new BPRecursiveAction(lpmSet, lpmOther, lpm, projection, projectionMapping, null,
						CombinationType.SEQ, gammaSetsMap, gammASetsMap, evaluatedLPMs, gammA, evaluator, params,
						context));
				subtasks.add(new BPRecursiveAction(lpmSet, lpm, lpmOther, projection, projectionMapping, null,
						CombinationType.LOOP, gammaSetsMap, gammASetsMap, evaluatedLPMs, gammA, evaluator, params,
						context));
				subtasks.add(new BPRecursiveAction(lpmSet, lpmOther, lpm, projection, projectionMapping, null,
						CombinationType.LOOP, gammaSetsMap, gammASetsMap, evaluatedLPMs, gammA, evaluator, params,
						context));
				subtasks.add(new BPRecursiveAction(lpmSet, lpmOther, lpm, projection, projectionMapping, null,
						CombinationType.AND, gammaSetsMap, gammASetsMap, evaluatedLPMs, gammA, evaluator, params,
						context));
			}
		}
//		XLog projectionLoopSelf = new XLogImpl(params.getDiscoveryLog().getAttributes());
//		Map<Integer, Integer> projectionMappingLoopSelf = new HashMap<Integer, Integer>();
//		Iterator<Integer> it = lpm.getProjectionMap().iterator();
//		int j = 0;
//		while (it.hasNext()) {
//			int traceIndex = it.next();
//			projectionLoopSelf.add(params.getDiscoveryLog().get(traceIndex));
//			projectionMappingLoopSelf.put(j, traceIndex);
//			j++;
//		}
//		subtasks.add(new LPMRecursiveAction(lpmSet, lpm, lpm, projectionLoopSelf, projectionMappingLoopSelf, null,
//				CombinationType.LOOP, gammaSetsMap, gammASetsMap, evaluatedLPMs, gammA, evaluator, params, context));

		return subtasks;
	}

	private List<BPRecursiveAction> createInfreqSubtasks(ProcessTreeSet lpmSet, BehavioralPattern lpm) {
//		System.out.println("Infreq Set of representative string : " + lpmSet.getRepresString());
//		System.out.println("Added LPM : " + lpm.getRepresString());

		List<BPRecursiveAction> subtasks = new ArrayList<BPRecursiveAction>();
		for (int i = 0; i < lpmSet.size(); i++) {
			BehavioralPattern lpmOther = lpmSet.get(i);
			Set<Integer> union = new HashSet<Integer>(lpm.getProjectionMap());
			union.addAll(lpmOther.getProjectionMap());

			subtasks.add(new BPRecursiveAction(lpmSet, lpm, lpmOther, null, null, union, CombinationType.XOR,
					gammaSetsMap, gammASetsMap, evaluatedLPMs, gammA, evaluator, params, context));
			subtasks.add(new BPRecursiveAction(lpmSet, lpmOther, lpm, null, null, union, CombinationType.XOR,
					gammaSetsMap, gammASetsMap, evaluatedLPMs, gammA, evaluator, params, context));

		}
		return subtasks;
	}

	private String represStringGen(List<Object> linearFormat, int positionNodeCombi) {

		String represString = "";
		for (int i = 0; i < linearFormat.size(); i++)
			if (i == positionNodeCombi)
				represString += "___,";
			else if (linearFormat.get(i) instanceof Leaf)
				represString += ((Leaf) linearFormat.get(i)).getName() + ",";
			else
				represString += ((Operator) linearFormat.get(i)).getName() + ",";
		return represString;
	}

	private void removeConstructionTree(BehavioralPattern lpm) {
		if (lpm.getSeed1() != null) {
			gammA.remove(lpm.getSeed1());
			gammA.remove(lpm.getSeed2());
			if (lpm.getSeed1().getDefiningOperator() != CombinationType.XOR)
				removeConstructionTree(lpm.getSeed1());
			if (lpm.getSeed2().getDefiningOperator() != CombinationType.XOR)
				removeConstructionTree(lpm.getSeed2());
		}
	}

}
