package org.processmining.behavioralpatternmining.algorithms;


import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.concurrent.ForkJoinPool;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.impl.XLogImpl;
import org.processmining.behavioralpatternmining.models.BehavioralPattern;
import org.processmining.behavioralpatternmining.models.BehavioralPatternSet;
import org.processmining.behavioralpatternmining.models.CombinationType;
import org.processmining.behavioralpatternmining.models.Leaf;
import org.processmining.behavioralpatternmining.models.ProcessTreeSet;
import org.processmining.behavioralpatternmining.parameters.BehavioralPatternParameters;
import org.processmining.framework.plugin.PluginContext;


public class BehavioralPatternMiner {

	public BehavioralPatternSet mine(PluginContext context, XLog eventLog, BehavioralPatternParameters params) {

		XLogInfo info = (new XLogInfoFactory()).createLogInfo(eventLog);

		//		List<LocalProcessModel> gammA1 = new ArrayList<LocalProcessModel>();
		List<BehavioralPattern> gamma1 = new ArrayList<BehavioralPattern>();
		//		Set<String> represStringGamma = new HashSet<String>();

		Map<String, ProcessTreeSet> gammASetsMap = Collections
				.synchronizedMap(new HashMap<String, ProcessTreeSet>());
		Map<String, ProcessTreeSet> gammaSetsMap = Collections
				.synchronizedMap(new HashMap<String, ProcessTreeSet>());
		Set<BehavioralPattern> gammA = Collections.synchronizedSet(new TreeSet<BehavioralPattern>());
		Map<String, BehavioralPattern> evaluatedLPMs = Collections
				.synchronizedMap(new HashMap<String, BehavioralPattern>());
		Map<Integer, String> activityMap = Collections
				.synchronizedMap(new HashMap<Integer, String>());
		Map<Integer, List<String>> activityXorMap = Collections
				.synchronizedMap(new HashMap<Integer, List<String>>());
		Map<String, Integer> intActivityMap = Collections
				.synchronizedMap(new HashMap<String, Integer>());

		
		ProcessTreeEvaluator evaluator = new ProcessTreeEvaluator(params, activityMap, activityXorMap, intActivityMap);

		ProcessTreeSet gammA1 = new ProcessTreeSet();
		gammA1.setDepth(0);
		gammA1.setPositionCombi(0);
		gammA1.setRepresString("___");
		gammA1.setContainsFreq(true);
		if (params.getDepthMax() >= 1)
			gammA1.setExtensible(true);

		XLog projectionInit = new XLogImpl(params.getDiscoveryLog().getAttributes());
		Map<Integer, Integer> projectionMappingInit = new HashMap<Integer, Integer>();
		for (int k = 0; k < params.getDiscoveryLog().size(); k++) {
			projectionInit.add(params.getDiscoveryLog().get(k));
			projectionMappingInit.put(k, k);
		}
		
		ForkJoinPool p = new ForkJoinPool(Runtime.getRuntime().availableProcessors());
		BPRecursiveAction rootTask = null;
		int activity = 0;
		for (XEventClass c : info.getNameClasses().getClasses()) {
			activityMap.put(activity, c.getId());
			intActivityMap.put(c.getId(), activity);
			Leaf man = new Leaf(activity);
			man.setDepth(0);
			man.setRepresLeaf(activity);
			
			List<Object> linearFormat = new ArrayList<Object>();
			linearFormat.add(man);
			
			BehavioralPattern lpm = evaluator.evaluateProcessTreeOnLog(context, man, projectionInit,
					projectionMappingInit, false);
			evaluatedLPMs.put(man.getName() + "", lpm);
			lpm.setRepresString(man.getName() + "");
			lpm.setLinearFormat(linearFormat);
			lpm.setDepth(0);
			lpm.setSeed1(null);
			lpm.setSeed2(null);
			lpm.setDefiningOperator(null);
			lpm.setRoot(man);
			if (lpm.getSupport() >= params.getSupportMinimum()) {
				gammA.add(lpm);
				if (gammA1.isExtensible() && gammA1.size() > 0) {
					rootTask = new BPRecursiveAction(gammA1, lpm, null, null, null, null, CombinationType.ROOT,
							gammASetsMap, gammaSetsMap, evaluatedLPMs, gammA, evaluator, params, context);
					p.invoke(rootTask);
				}

				gammA1.add(lpm);
			} else {
				gamma1.add(lpm);
			}
			activity++;
		}

		int i = 0, j;

		while (i < gamma1.size() - 1) {
			BehavioralPattern activity1 = gamma1.get(i);
			j = i + 1;
			while (j < gamma1.size()) {

				BehavioralPattern activity2 = gamma1.get(j);

				Set<Integer> projectionChoice = new HashSet<Integer>(activity1.getProjectionMap());
				projectionChoice.addAll(activity2.getProjectionMap());

				if (projectionChoice.size() * 1.0 / params.getDiscoveryLog().size() >= params.getSupportMinimum()) {
					List<String> activityXor = new ArrayList<String>();
					
					String activity1Name = activityMap.get(((Leaf) activity1.getRoot()).getName());
					String activity2Name = activityMap.get(((Leaf) activity2.getRoot()).getName());
					activityXor.add(activity1Name);
					activityXor.add(activity2Name);
					
					Leaf man = new Leaf(activity);
					man.setDepth(0);
					man.setRepresLeaf(activity);
					activityXorMap.put(activity, activityXor);
					BehavioralPattern lpmChoice = new BehavioralPattern();
					lpmChoice.setProjectionMap(projectionChoice);
					lpmChoice.setFrequency(projectionChoice.size(), params.getDiscoveryLog().size());
					Set<List<Integer>> language = new HashSet<List<Integer>>();
					language.add(new ArrayList<Integer>(Arrays.asList(((Leaf) activity1.getRoot()).getName())));
					language.add(new ArrayList<Integer>(Arrays.asList(((Leaf) activity2.getRoot()).getName())));
					lpmChoice.setLanguage(language);
					lpmChoice.setLanguageSeen(language);
					lpmChoice.setLanguageRatio(1.0);
					List<Object> linearFormat = new ArrayList<Object>();
					linearFormat.add(man);
					evaluatedLPMs.put(man.getName() + "", lpmChoice);
					lpmChoice.setRepresString(man.getName() + "");
					lpmChoice.setDepth(0);
					lpmChoice.setLinearFormat(linearFormat);
					lpmChoice.setSeed1(null);
					lpmChoice.setSeed2(null);
					lpmChoice.setDefiningOperator(null);
					lpmChoice.setRoot(man);
					if (gammA1.isExtensible() && gammA1.size() > 0) {
						rootTask = new BPRecursiveAction(gammA1, lpmChoice, null, null, null, null,
								CombinationType.ROOT, gammASetsMap, gammaSetsMap, evaluatedLPMs, gammA, evaluator,
								params, context);
						p.invoke(rootTask);
					}
					gammA1.add(lpmChoice);
					activity++;
				}
				j++;
			}
			i++;
		}

//		BufferedWriter writer;
//		try {
//			writer = new BufferedWriter(new FileWriter("behavior patterns.txt"));
//			synchronized(gammA)
//			{
//				Iterator<BehavioralPattern> it = gammA.iterator();
//				
//				while(it.hasNext())
//				{
//					BehavioralPattern lpm = it.next();
//					//System.out.println(lpm.finalPrint(activityMap, activityXorMap));
//					writer.write(lpm.finalPrint(activityMap, activityXorMap)+System.lineSeparator()+System.lineSeparator());
//				}
//			}
//		    writer.close();
//		} catch (IOException e) {
//			// TODO Auto-generated catch block
//			e.printStackTrace();
//		}
		
		BehavioralPatternSet bPSet = new BehavioralPatternSet();
		bPSet.setActivityMap(activityMap);
		bPSet.setActivityXorMap(activityXorMap);
		synchronized(gammA)
		{
			Iterator<BehavioralPattern> it = gammA.iterator();
			
			while(it.hasNext())
			{
				BehavioralPattern lpm = it.next();
				//System.out.println(lpm.finalPrint(activityMap, activityXorMap));
				bPSet.add(lpm);
			}
		}

		return bPSet;
	}
}
