package org.processmining.behavioralpatternmining.algorithms;


import static org.processmining.lpm.util.AcceptingPetrinetUtils.calculateLanguage;
import static org.processmining.lpm.util.PetrinetUtils.DUMMY;
import static org.processmining.lpm.util.PetrinetUtils.projectLogOnNetAlphabet;
import static org.processmining.lpm.util.PetrinetUtils.transitionToEventClassMapperByLabel;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.acceptingpetrinet.models.AcceptingPetriNet;
import org.processmining.behavioralpatternmining.models.BehavioralPattern;
import org.processmining.behavioralpatternmining.models.Node;
import org.processmining.behavioralpatternmining.parameters.BehavioralPatternParameters;
import org.processmining.behavioralpatternmining.util.ProcessTreeUtils;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.lpm.adjustedalignments.CostBasedCompleteParam;
import org.processmining.lpm.adjustedalignments.NBestOptAlignmentsNoModelMoveGraphSamplingAlg;
import org.processmining.lpm.util.PetrinetUtils;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;
import org.processmining.plugins.petrinet.replayer.matchinstances.PNLogMatchInstancesReplayer;
import org.processmining.plugins.petrinet.replayresult.PNMatchInstancesRepResult;
import org.processmining.plugins.petrinet.replayresult.StepTypes;
import org.processmining.plugins.replayer.replayresult.AllSyncReplayResult;

import nl.tue.astar.AStarException;

public class ProcessTreeEvaluator {

	private BehavioralPatternParameters params;
	private Map<Integer, String> activityMap;
	private Map<Integer, List<String>> activityXorMap;
	private Map<String, Integer> intActivityMap;

	public ProcessTreeEvaluator() {
	}

	public ProcessTreeEvaluator(BehavioralPatternParameters params, Map<Integer, String> activityMap, Map<Integer, List<String>> activityXorMap, Map<String, Integer> intActivityMap) {
		this.params = params;
		this.activityMap = activityMap;
		this.activityXorMap = activityXorMap;
		this.intActivityMap = intActivityMap;
	}

	public BehavioralPattern evaluateProcessTreeOnLog(PluginContext context, Node root, XLog projection, Map<Integer, Integer> projectionMapping, boolean verbose) {
		AcceptingPetriNet apn = ProcessTreeUtils.processTreeToAPN(context, root, activityMap, activityXorMap);
		if(apn == null) return null;
		
		if(verbose) System.out.println(PetrinetUtils.prettyPrintPetrinet(apn.getNet()));
		
		int maxLoop = 1;
		Set<List<String>> petriNetLanguage = calculateLanguage(apn, maxLoop, true);
		Set<List<Integer>> language = new HashSet<List<Integer>>();
		Iterator<List<String>> it1 = petriNetLanguage.iterator();
		while(it1.hasNext()) {
			List<Integer> wordLanguage = new ArrayList<Integer>();
			language.add(wordLanguage);
			List<String> word = it1.next();
			for(int i = 0; i < word.size(); i++)
				wordLanguage.add(intActivityMap.get(word.get(i)));
			
		}
		Set<List<String>> languageElementsInLog = new HashSet<List<String>>();
		Petrinet pn = apn.getNet();
		

		XLog projectedLog = null;
		int maxTraceLength = 0;

		Object[] o = projectLogOnNetAlphabet(projection, pn);
		projectedLog = (XLog) o[0];
		maxTraceLength = (int) o[2];
		Set<Integer> projectionMap = new HashSet<Integer>();

		PNMatchInstancesRepResult alignments = null;
		try {
			alignments = calculateNAlignment(context, projectedLog, apn, maxTraceLength, params, verbose);
		} catch (NullPointerException e) {
		}

		if (alignments != null) {
			Iterator<AllSyncReplayResult> sync = alignments.iterator();
			while (sync.hasNext()) {
				AllSyncReplayResult srr = sync.next();
				if(verbose) System.out.println(srr.getInfo());
				boolean counted = false;

				List<Object> nodeInstances = srr.getNodeInstanceLst().get(0);
				List<StepTypes> stepTypes = srr.getStepTypesLst().get(0);

				List<String> languageElem = new ArrayList<String>();

				for (int i = 0; i < nodeInstances.size(); i++) {
					Object nodeInstance = nodeInstances.get(i);
					StepTypes st = stepTypes.get(i);
					if (nodeInstance instanceof Transition) {
						if (!counted && st.name().equals("LMGOOD")) {
							Iterator<Integer> it = srr.getTraceIndex().iterator();
							while(it.hasNext())
							{
								int traceIndex = it.next();
								projectionMap.add(projectionMapping.get(traceIndex));
							}
							counted = true;
						}

						Transition transition = (Transition) nodeInstance;

						if (!transition.isInvisible()) {
							languageElem.add(transition.getLabel());
							if (verbose) {
								System.out.println("t: " + transition.getLabel() + " __ " + st.name());
							}
						} 
					} else {
						if (verbose) {
							XEventClass eventClass = (XEventClass) nodeInstance;
							System.out.println("e: " + eventClass + " __ " + st.name());
						}
					}
				}
				if(petriNetLanguage.contains(languageElem))
				languageElementsInLog.add(languageElem);
			}
		}


		
		
		Set<List<Integer>> languageSeen = new HashSet<List<Integer>>();
		Iterator<List<String>> it2 = languageElementsInLog.iterator();
		while(it2.hasNext()) {
			List<Integer> wordLanguage = new ArrayList<Integer>();
			languageSeen.add(wordLanguage);
			List<String> word = it2.next();
			for(int i = 0; i < word.size(); i++)
				wordLanguage.add(intActivityMap.get(word.get(i)));
			
		}
		
		double languageElementsSeenRatio = ((double) languageSeen.size()) / language.size();

		BehavioralPattern scoredAPN = new BehavioralPattern();
		scoredAPN.setFrequency(projectionMap.size(), params.getDiscoveryLog().size());
		scoredAPN.setLanguageRatio(languageElementsSeenRatio);
		scoredAPN.setLanguage(language);
		scoredAPN.setLanguageSeen(languageSeen);
		scoredAPN.setProjectionMap(projectionMap);


		return scoredAPN;
	}

	public static PNMatchInstancesRepResult calculateNAlignment(PluginContext context, XLog log, AcceptingPetriNet apn,
			int maxTraceLength, BehavioralPatternParameters params, boolean verbose) {

		Petrinet ptnet = apn.getNet();
		TransEvClassMapping oldMap = transitionToEventClassMapperByLabel(log, ptnet);
		XLogInfo logInfo = XLogInfoFactory.createLogInfo(log, oldMap.getEventClassifier());

		CostBasedCompleteParam parameter = new CostBasedCompleteParam(logInfo.getEventClasses().getClasses(), DUMMY,
				ptnet.getTransitions(), Integer.MAX_VALUE, 1);

		parameter.getMapEvClass2Cost().remove(DUMMY);
		parameter.getMapEvClass2Cost().put(DUMMY, 1);
		parameter.setGUIMode(false);

		for (Transition t : apn.getNet().getTransitions()) {
			if (!t.isInvisible())
				parameter.getMapTrans2Cost().put(t, maxTraceLength);
			else
				parameter.getMapTrans2Cost().put(t, 0);
		}

		parameter.setCreateConn(false);

		Marking initialMarking = apn.getInitialMarking();
		Marking finalMarking = new ArrayList<Marking>(apn.getFinalMarkings()).get(0);

		parameter.setMaxNumOfStates(
				params.getAlignmentMaxNumStatesPerTransition() * apn.getNet().getTransitions().size());
		parameter.setInitialMarking(initialMarking);
		parameter.setFinalMarkings(finalMarking);
		if (verbose) {
			System.out.println("log: ");
			for (XTrace trace : log) {
				for (XEvent event : trace) {
					System.out.print(event.getAttributes().get("concept:name") + ", ");
				}
				System.out.println();
			}

			System.out.println();
			System.out.println("mapping:                        " + oldMap);
			System.out.println("parameter.getMapTrans2Cost():   " + parameter.getMapTrans2Cost());
			System.out.println("parameter.getMapEvClass2Cost(): " + parameter.getMapEvClass2Cost());
		}

		Object[] parameters = new Object[] { parameter.getMapTrans2Cost(), parameter.getMaxNumOfStates(),
				parameter.getMapEvClass2Cost(), 1 };
		// select algorithm without ILP
		PNLogMatchInstancesReplayer replayer = new PNLogMatchInstancesReplayer();
		NBestOptAlignmentsNoModelMoveGraphSamplingAlg alg = new NBestOptAlignmentsNoModelMoveGraphSamplingAlg();

		PNMatchInstancesRepResult pnRepResult = null;
		try {
			pnRepResult = replayer.replayLog(context, ptnet, log, oldMap, initialMarking, finalMarking, alg,
					parameters);
		} catch (AStarException e) {
			e.printStackTrace();
		} catch (Throwable e) {
			e.printStackTrace();
		}
		
		return pnRepResult;
	}

}