package org.processmining.behavioralpatternmining.dialogs;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;

import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSpinner;
import javax.swing.SpinnerModel;
import javax.swing.SpinnerNumberModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import org.processmining.behavioralpatternmining.parameters.BehavioralPatternParameters;

public class ConfigParametersDialog extends JPanel implements ChangeListener {

	private static final long serialVersionUID = 1L;
	
	private BehavioralPatternParameters params;
	private JSpinner supportSpinner;
	private JSpinner languageSpinner;
	private JSpinner depthSpinner;
	
	public ConfigParametersDialog(BehavioralPatternParameters params)
	{
		this.params = params;
		Dimension preferredSize = new Dimension(100, 20);
		
		this.setLayout(new GridBagLayout());
		this.setAlignmentX(Component.LEFT_ALIGNMENT);
		this.setOpaque(false);
		
		GridBagConstraints c = new GridBagConstraints();
		
		JLabel supportLabel = new JLabel("Support threshold : ");
		c.gridx = 0;
		c.gridy = 0;
		c.anchor = GridBagConstraints.LINE_END;
		c.insets = new Insets(10, 0, 0, 0);
		this.add(supportLabel, c);
		
		SpinnerModel sM = new SpinnerNumberModel(params.getSupportMinimum(), 0.0, 1.0, 0.05);
		supportSpinner = new JSpinner(sM);
		supportSpinner.setPreferredSize(preferredSize);
		supportSpinner.setMaximumSize(preferredSize);
		c.gridx = 1;
		c.gridy = 0;
		c.anchor = GridBagConstraints.LINE_START;
		c.insets = new Insets(10, 0, 0, 0);
		this.add(supportSpinner, c);
		supportSpinner.addChangeListener(this);
		
		JLabel languageLabel = new JLabel("Language fitness threshold : ");
		c.gridx = 0;
		c.gridy = 1;
		c.anchor = GridBagConstraints.LINE_END;
		this.add(languageLabel, c);
		
		SpinnerModel sML = new SpinnerNumberModel(params.getLanguageFitMinimum(), 0.0, 1.0, 0.05);
		languageSpinner = new JSpinner(sML);
		languageSpinner.setPreferredSize(preferredSize);
		languageSpinner.setMaximumSize(preferredSize);
		c.gridx = 1;
		c.gridy = 1;
		c.anchor = GridBagConstraints.LINE_START;
		this.add(languageSpinner, c);
		languageSpinner.addChangeListener(this);


		
		JLabel depthLabel = new JLabel("Depth : ");
		c.gridx = 0;
		c.gridy = 2;
		c.anchor = GridBagConstraints.LINE_END;
		this.add(depthLabel, c);
		
		SpinnerModel sMD = new SpinnerNumberModel(params.getDepthMax(), 1, Integer.MAX_VALUE, 1);
		depthSpinner = new JSpinner(sMD);
		depthSpinner.setPreferredSize(preferredSize);
		depthSpinner.setMaximumSize(preferredSize);
		c.gridx = 1;
		c.gridy = 2;
		c.anchor = GridBagConstraints.LINE_START;
		this.add(depthSpinner, c);
		depthSpinner.addChangeListener(this);

		
	}

	public void stateChanged(ChangeEvent e) {
		if(e.getSource().equals(supportSpinner)) {
			params.setSupportMinimum((Double) supportSpinner.getValue());
		} else if(e.getSource().equals(languageSpinner)) {
			params.setLanguageFitMinimum((Double) languageSpinner.getValue());
		} else if(e.getSource().equals(depthSpinner)) {
			params.setDepthMax((Integer) depthSpinner.getValue());
			System.out.println(params.getDepthMax());
		}
		
	}
	
	
	
}

