package org.processmining.behavioralpatternmining.models;


import java.util.List;
import java.util.Map;
import java.util.Set;

public class BehavioralPattern implements Comparable<BehavioralPattern> {

	private int frequency;
	private double support;
	private double languageRatio;
	private Set<List<Integer>> language;
	private Set<List<Integer>> languageSeen;
	private Set<Integer> projectionMap;
	private List<Object> linearFormat;
	private String represString;
	private int depth;
	private BehavioralPattern seed1;
	private BehavioralPattern seed2;
	private CombinationType definingOperator;
	private Node root;

	public BehavioralPattern(int frequency, double languageRatio, Set<List<Integer>> language,
			Set<List<Integer>> languageSeen) {
		this.frequency = frequency;
		this.languageRatio = languageRatio;
		this.language = language;
		this.languageSeen = languageSeen;
	}

	public BehavioralPattern() {
		// TODO Auto-generated constructor stub
	}

	//	public Map<String, Integer> getLogActivityCountMap(){
	//		return logActivityCountMap;
	//	}
	//	
	//	public void setLogActivityCountMap(Map<String, Integer> logActivityCountsMap){
	//		this.logActivityCountMap = logActivityCountsMap;
	//	}

	public int getFrequency() {
		return frequency;
	}

	public double getSupport() {
		return support;
	}

	public int compareTo(BehavioralPattern o) {
		int hashComparison = Integer.compare(hashCode(), o.hashCode());
		if (hashComparison == 0) // equal nets
			return 0;
		int expression = frequency - o.getFrequency();
		int result = expression > 0 ? -1 : expression < 0 ? 1 : 0;
		if (result == 0) {
			double expression1 = languageRatio - o.getLanguageRatio();
			result = expression1 > 0 ? -1 : expression1 < 0 ? 1 : 0;
			if (result == 0)
				return hashComparison;
		}

		return result;
	}

	@Override
	public int hashCode() {
		return represString.hashCode();
	}

	@Override
	public boolean equals(Object other) {
		if (!(other instanceof BehavioralPattern))
			return false;
		BehavioralPattern apnOther = (BehavioralPattern) other;
		return represString.equals(apnOther.getRepresString());
	}

	public Set<List<Integer>> getLanguage() {
		return language;
	}

	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append("Process tree:           " + processTree());
		sb.append(System.lineSeparator());
		sb.append("Representative String:           " + represString);
		sb.append(System.lineSeparator());
		sb.append("Freq :          " + getFrequency());
		sb.append(System.lineSeparator());
		sb.append("Support :          " + getSupport());
		sb.append(System.lineSeparator());
		sb.append("Language ratio:          " + getLanguageRatio());
		sb.append(System.lineSeparator());
		sb.append("Language:                " + language);
		sb.append(System.lineSeparator());
		sb.append("Language seen:           " + languageSeen);
		sb.append(System.lineSeparator());

		return sb.toString();
	}

	public Set<List<Integer>> getLanguageSeen() {
		return languageSeen;
	}

	public double getLanguageRatio() {
		return languageRatio;
	}

	public void setLanguageRatio(double languageRatio) {
		this.languageRatio = languageRatio;
	}

	public void setFrequency(int frequency, int logSize) {
		this.frequency = frequency;

		support = frequency * 1.0 / logSize;
	}

	public void setLanguage(Set<List<Integer>> language) {
		this.language = language;
	}

	public void setLanguageSeen(Set<List<Integer>> languageSeen) {
		this.languageSeen = languageSeen;
	}

	public void setRepresString(String represString) {
		this.represString = represString;
	}

	public Set<Integer> getProjectionMap() {
		return projectionMap;
	}

	public void setProjectionMap(Set<Integer> projectionMap) {
		this.projectionMap = projectionMap;
	}

	public List<Object> getLinearFormat() {
		return linearFormat;
	}

	public void setLinearFormat(List<Object> linearFormat) {
		this.linearFormat = linearFormat;
	}

	public String getRepresString() {
		return represString;
	}

	public int getDepth() {
		return depth;
	}

	public void setDepth(int depth) {
		this.depth = depth;
	}

	public BehavioralPattern getSeed1() {
		return seed1;
	}

	public void setSeed1(BehavioralPattern seed1) {
		this.seed1 = seed1;
	}

	public BehavioralPattern getSeed2() {
		return seed2;
	}

	public void setSeed2(BehavioralPattern seed2) {
		this.seed2 = seed2;
	}

	public CombinationType getDefiningOperator() {
		return definingOperator;
	}

	public void setDefiningOperator(CombinationType definingOperator) {
		this.definingOperator = definingOperator;
	}

	public Node getRoot() {
		return root;
	}

	public void setRoot(Node root) {
		this.root = root;
	}

	private String processTree() {
		StringBuilder sb = new StringBuilder();
		construProcessTreeReq(sb, root);
		return sb.toString();
	}

	private void construProcessTreeReq(StringBuilder sb, Node node) {
		if (node instanceof Leaf)
			sb.append(((Leaf) node).getName());
		else {
			Operator op = (Operator) node;
			sb.append("(");
			construProcessTreeReq(sb, op.getChild1());
			sb.append(op.getName());
			construProcessTreeReq(sb, op.getChild2());
			sb.append(")");
		}
	}

	public String finalPrint(Map<Integer, String> activityMap, Map<Integer, List<String>> activityXorMap) {
		StringBuilder sb = new StringBuilder();
		sb.append("Process tree:           " + finalProcessTree(activityMap, activityXorMap));
		sb.append(System.lineSeparator());
		sb.append("Representative String:           " + represString);
		sb.append(System.lineSeparator());
		sb.append("Freq :          " + getFrequency());
		sb.append(System.lineSeparator());
		sb.append("Support :          " + getSupport());
		sb.append(System.lineSeparator());
		sb.append("Language ratio:          " + getLanguageRatio());
		sb.append(System.lineSeparator());
		sb.append("Language:                " + language);
		sb.append(System.lineSeparator());
		sb.append("Language seen:           " + languageSeen);
		sb.append(System.lineSeparator());

		return sb.toString();
	}

	private String finalProcessTree(Map<Integer, String> activityMap, Map<Integer, List<String>> activityXorMap) {
		StringBuilder sb = new StringBuilder();
		construFinalProcessTreeReq(sb, root, activityMap, activityXorMap);
		return sb.toString();
	}

	private void construFinalProcessTreeReq(StringBuilder sb, Node node, Map<Integer, String> activityMap,
			Map<Integer, List<String>> activityXorMap) {
		if (node instanceof Leaf) {
			int name = ((Leaf) node).getName();
			if(activityMap.get(name) != null) sb.append(activityMap.get(name));
			else if (activityXorMap.get(name) != null) 
			{
				List<String> xorAct = activityXorMap.get(name);
				sb.append("(" + xorAct.get(0) + " xor " + xorAct.get(1) + ")");
			}
		} else {
			Operator op = (Operator) node;
			sb.append("(");
			construFinalProcessTreeReq(sb, op.getChild1(), activityMap, activityXorMap);
			sb.append(" " + op.getName() + " ");
			construFinalProcessTreeReq(sb, op.getChild2(), activityMap, activityXorMap);
			sb.append(")");
		}
	}
}
