package org.processmining.behavioralpatternmining.plugins;


import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.model.XLog;
import org.processmining.behavioralpatternmining.algorithms.BehavioralPatternMiner;
import org.processmining.behavioralpatternmining.dialogs.ConfigParametersDialog;
import org.processmining.behavioralpatternmining.models.BehavioralPatternSet;
import org.processmining.behavioralpatternmining.parameters.BehavioralPatternParameters;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginLevel;
import org.processmining.framework.plugin.annotations.PluginVariant;



@Plugin(name = "Mine Behavioral Patterns with COBPAM",
		level = PluginLevel.PeerReviewed,
		returnLabels = { "Behavioral Patterns Set" }, 
		returnTypes = {	BehavioralPatternSet.class },
		categories = { PluginCategory.Discovery, PluginCategory.Analytics },
		parameterLabels = { "XLog" },
		keywords = { "Frequent process trees discovery", "Behavioral patterns discovery"},
		help = "Discover frequent fragments of behavior from event logs.", 
		userAccessible = true
		)

public class CobpamPlugin {

	private BehavioralPatternMiner miner = new BehavioralPatternMiner();

	@UITopiaVariant(
			affiliation = UITopiaVariant.EHV, 
			author = "M. Acheli", 
			email = "mehdi.acheli@dauphine.fr",
			uiLabel = UITopiaVariant.USEVARIANT,
			pack = "COBPAM"
		)
	@PluginVariant(
			variantLabel = "COBPAM - config dialog", 
			requiredParameterLabels = { 0 }
		)
	public BehavioralPatternSet mine(UIPluginContext context, XLog log) {
		BehavioralPatternSet bPSet= null;
		BehavioralPatternParameters params = new BehavioralPatternParameters();
		params.setDiscoveryLog(log);
		InteractionResult result = context.showConfiguration("Choose configuration", new ConfigParametersDialog(params));
		if (result != InteractionResult.CONTINUE) {
			context.getFutureResult(0).cancel(true);
			return null;
		}
		bPSet = miner.mine(context, log, params);
		return bPSet;
	}

	
}

