package org.processmining.behavioralpatternmining.plugins;


import java.awt.Color;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.util.List;

import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;

import org.processmining.behavioralpatternmining.models.BehavioralPattern;
import org.processmining.behavioralpatternmining.models.BehavioralPatternSet;
import org.processmining.behavioralpatternmining.models.Leaf;
import org.processmining.behavioralpatternmining.models.Node;
import org.processmining.behavioralpatternmining.models.Operator;
import org.processmining.contexts.uitopia.annotations.Visualizer;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginLevel;
import org.processmining.processtree.Block;
import org.processmining.processtree.Originator.Resource;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.Task.Automatic;
import org.processmining.processtree.Task.Manual;
import org.processmining.processtree.impl.AbstractBlock;
import org.processmining.processtree.impl.AbstractOriginator;
import org.processmining.processtree.impl.AbstractTask;
import org.processmining.processtree.impl.ProcessTreeImpl;
import org.processmining.processtree.visualization.tree.TreeVisualization;

public class VisualizeBehaviorPatterns {
	@Plugin(name = "Show behavioral patterns", level = PluginLevel.PeerReviewed, returnLabels = {
			"Visualization of behavioral patterns " }, returnTypes = {
					JComponent.class }, parameterLabels = { "Behavioral Patterns Set" }, userAccessible = true)
	@Visualizer
	public JComponent visualize(PluginContext context, BehavioralPatternSet bPSet) {
		TreeVisualization visualizer = new TreeVisualization();
		JTabbedPane tabbedPane = new JTabbedPane();
		for (int i = 0; i < bPSet.size(); i++) {
			BehavioralPattern behaviorPattern = bPSet.get(i);
			JPanel wrapper = new JPanel();
			String label = "Tree " + (i + 1);
			tabbedPane.add(label, wrapper);
			wrapper.setLayout(new GridBagLayout());
			GridBagConstraints c = new GridBagConstraints();
			c.gridx = 0;
			c.gridy = 0;
			JComponent treeVis = visualizer.visualize(context, createProcessTree(behaviorPattern, bPSet));
			//treeVis.setPreferredSize(new Dimension((int) treeVis.getPreferredSize().getWidth(), (int) tabbedPane.getPreferredSize().getHeight()));
			wrapper.add(treeVis, c);
			JLabel labelTree = new JLabel("<html>"
					+ "<p><strong>Quality Metrics</strong><p>"
					+ "<ul>"
					+ "<li><b>Support</b> : " + behaviorPattern.getSupport() +"</li>"
					+ "<li><b>Language Fitness</b> : " + behaviorPattern.getLanguageRatio() +"</li>"
					+ "</ul>"
					+ "</html>");
			c.gridx = 1;
			c.gridy = 0;
			c.insets = new Insets(0, 10, 0, 0);
			wrapper.add(labelTree, c);
			wrapper.setBackground(new Color(255, 255, 255));
			
		}
		return tabbedPane;
	}

	private ProcessTree createProcessTree(BehavioralPattern bP, BehavioralPatternSet bPSet) {
		ProcessTree p = new ProcessTreeImpl();
		Resource res = new AbstractOriginator.Resource("res");
		p.addOriginator(res);
		recursConstruProcessTree(p, bP.getRoot(), null, res, bPSet);
		return p;
	}

	private void recursConstruProcessTree(ProcessTree p, Node n, Block parent, Resource res,
			BehavioralPatternSet bPSet) {
		if (n instanceof Leaf) {
			if (bPSet.getActivityMap().containsKey(((Leaf) n).getName())) {
				String nameTask;
				nameTask = bPSet.getActivityMap().get(((Leaf) n).getName());
				Manual task = new AbstractTask.Manual(nameTask, res);
				p.addNode(task);
				if (parent != null) {
					parent.addChild(task);
				} else
					p.setRoot(task);
			} else 
			{
				List<String> nameXorTasks = bPSet.getActivityXorMap().get(((Leaf) n).getName());
				Manual task1 = new AbstractTask.Manual(nameXorTasks.get(0), res);
				p.addNode(task1);
				Manual task2 = new AbstractTask.Manual(nameXorTasks.get(1), res);
				p.addNode(task2);
				Block operator = new AbstractBlock.Xor("Xor");
				p.addNode(operator);
				operator.addChild(task1);
				operator.addChild(task2);
				if (parent != null) {
					parent.addChild(operator);
				} else
					p.setRoot(operator);
			}
			
		} else if (n instanceof Operator && ((Operator) n).getName().equals("Seq")) {
			Block operator = new AbstractBlock.Seq("Seq");
			p.addNode(operator);
			if (parent != null) {
				parent.addChild(operator);
			} else
				p.setRoot(operator);
			recursConstruProcessTree(p, ((Operator) n).getChild1(), operator, res, bPSet);
			recursConstruProcessTree(p, ((Operator) n).getChild2(), operator, res, bPSet);
		} else if (n instanceof Operator && ((Operator) n).getName().equals("And")) {
			Block operator = new AbstractBlock.And("And");
			p.addNode(operator);
			if (parent != null) {
				parent.addChild(operator);
			} else
				p.setRoot(operator);
			recursConstruProcessTree(p, ((Operator) n).getChild1(), operator, res, bPSet);
			recursConstruProcessTree(p, ((Operator) n).getChild2(), operator, res, bPSet);
		} else if (n instanceof Operator && ((Operator) n).getName().equals("Xor")) {
			Block operator = new AbstractBlock.Xor("Xor");
			p.addNode(operator);
			if (parent != null) {
				parent.addChild(operator);
			} else
				p.setRoot(operator);
			recursConstruProcessTree(p, ((Operator) n).getChild1(), operator, res, bPSet);
			recursConstruProcessTree(p, ((Operator) n).getChild2(), operator, res, bPSet);
		} else if (n instanceof Operator && ((Operator) n).getName().equals("Loop")) {
			Block operator = new AbstractBlock.XorLoop("Loop");
			p.addNode(operator);
			if (parent != null) {
				parent.addChild(operator);
			} else
				p.setRoot(operator);
			recursConstruProcessTree(p, ((Operator) n).getChild1(), operator, res, bPSet);
			recursConstruProcessTree(p, ((Operator) n).getChild2(), operator, res, bPSet);
			Automatic tau = new AbstractTask.Automatic("tau");
			p.addNode(tau);
			operator.addChild(tau);
		}
	}

}

