package org.processmining.behavioralpatternmining.util;


import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Deque;
import java.util.List;
import java.util.Map;
import java.util.concurrent.atomic.AtomicInteger;

import org.processmining.acceptingpetrinet.models.AcceptingPetriNet;
import org.processmining.acceptingpetrinet.models.impl.AcceptingPetriNetImpl;
import org.processmining.acceptingpetrinetclassicalreductor.parameters.ReduceUsingMurataRulesParameters;
import org.processmining.acceptingpetrinetclassicalreductor.plugins.ReduceUsingMurataRulesPlugin;
import org.processmining.behavioralpatternmining.models.Leaf;
import org.processmining.behavioralpatternmining.models.Node;
import org.processmining.behavioralpatternmining.models.Operator;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinet.impl.PetrinetImpl;
import org.processmining.models.semantics.petrinet.Marking;


public final class ProcessTreeUtils {

	static public String represString(Node root) {
		StringBuilder represString = new StringBuilder();
		construRepresStringRec(represString, root);
		return represString.toString();
	}

	private static void construRepresStringRec(StringBuilder represString, Node node) {
		if (node instanceof Leaf)
			represString.append(((Leaf) node).getName() + " ");
		else {
			Operator op = (Operator) node;
			if (op.getName().equals("Seq") || op.getName().equals("Loop")
					|| op.getChild1().getRepresLeaf() < op.getChild2().getRepresLeaf()) {
				construRepresStringRec(represString, op.getChild1());
				construRepresStringRec(represString, op.getChild2());
			} else {
				construRepresStringRec(represString, op.getChild2());
				construRepresStringRec(represString, op.getChild1());
			}
			represString.append(op.getName() + " ");
		}
	}

	static public AcceptingPetriNet processTreeToAPN(PluginContext context, Node root, Map<Integer, String> activityMap,
			Map<Integer, List<String>> activityXorMap) {
		AtomicInteger i = new AtomicInteger(0);
		Petrinet p = new PetrinetImpl("");
		List<Place> startEndPlaces = construPetriNetRec(p, root, activityMap, activityXorMap, i);
		Place pSource = startEndPlaces.get(0);
		Place pSink = startEndPlaces.get(1);

		Marking initialMarking = new Marking();
		initialMarking.add(pSource);
		Marking finalMarking = new Marking();
		finalMarking.add(pSink);
		AcceptingPetriNet ap = new AcceptingPetriNetImpl(p, initialMarking, finalMarking);

		ReduceUsingMurataRulesPlugin murata = new ReduceUsingMurataRulesPlugin();
		ReduceUsingMurataRulesParameters murataParams = new ReduceUsingMurataRulesParameters();
		murataParams.setRetainBehavior(true);
		murataParams.setTryConnections(false);
		ap = murata.run(context, ap, murataParams);
		return ap;
	}

	static List<Place> construPetriNetRec(Petrinet p, Node node, Map<Integer, String> activityMap,
			Map<Integer, List<String>> activityXorMap, AtomicInteger i) {
		List<Place> startEndPlaces = new ArrayList<Place>();
		if (node instanceof Leaf) {
			Leaf leaf = (Leaf) node;
			Place pSource = p.addPlace("source" + i);
			Place pSink = p.addPlace("sink" + i);
			if (activityMap.containsKey(leaf.getName())) {

				Transition t = p.addTransition(activityMap.get(leaf.getName()));
				p.addArc(pSource, t);
				p.addArc(t, pSink);
			} else if (activityXorMap.containsKey(leaf.getName())) {
				List<String> xorActivities = activityXorMap.get(leaf.getName());
				Transition t1 = p.addTransition(xorActivities.get(0));
				Transition t2 = p.addTransition(xorActivities.get(1));
				p.addArc(pSource, t1);
				p.addArc(pSource, t2);
				p.addArc(t1, pSink);
				p.addArc(t2, pSink);
			}
			startEndPlaces.add(pSource);
			startEndPlaces.add(pSink);
		} else if (node instanceof Operator && ((Operator) node).getName().equals("Seq")) {
			i.incrementAndGet();
			List<Place> node1 = construPetriNetRec(p, ((Operator) node).getChild1(), activityMap, activityXorMap, i);
			i.incrementAndGet();
			List<Place> node2 = construPetriNetRec(p, ((Operator) node).getChild2(), activityMap, activityXorMap, i);

			Place pSinkNode2 = node2.get(1);
			Place pSourceNode2 = node2.get(0);
			Place pSinkNode1 = node1.get(1);
			Place pSourceNode1 = node1.get(0);
			Transition tMSeq = p.addTransition("tau seq" + i);
			tMSeq.setInvisible(true);
			p.addArc(pSinkNode1, tMSeq);
			p.addArc(tMSeq, pSourceNode2);

			Transition tSSeq = p.addTransition("tau start seq" + i);
			tSSeq.setInvisible(true);
			p.addArc(tSSeq, pSourceNode1);

			Transition tESeq = p.addTransition("tau end seq" + i);
			tESeq.setInvisible(true);
			p.addArc(pSinkNode2, tESeq);

			Place pSource = p.addPlace("source" + i);
			Place pSink = p.addPlace("sink" + i);
			p.addArc(pSource, tSSeq);
			p.addArc(tESeq, pSink);
			startEndPlaces.add(pSource);
			startEndPlaces.add(pSink);
		} else if (node instanceof Operator && ((Operator) node).getName().equals("And")) {
			i.incrementAndGet();
			List<Place> node1 = construPetriNetRec(p, ((Operator) node).getChild1(), activityMap, activityXorMap, i);
			i.incrementAndGet();
			List<Place> node2 = construPetriNetRec(p, ((Operator) node).getChild2(), activityMap, activityXorMap, i);

			Place pSinkNode2 = node2.get(1);
			Place pSourceNode2 = node2.get(0);
			Place pSinkNode1 = node1.get(1);
			Place pSourceNode1 = node1.get(0);

			Transition tSAnd = p.addTransition("tau start and" + i);
			tSAnd.setInvisible(true);
			p.addArc(tSAnd, pSourceNode1);
			p.addArc(tSAnd, pSourceNode2);

			Transition tEAnd = p.addTransition("tau end and" + i);
			tEAnd.setInvisible(true);
			p.addArc(pSinkNode1, tEAnd);
			p.addArc(pSinkNode2, tEAnd);
			Place pSource = p.addPlace("source" + i);
			Place pSink = p.addPlace("sink" + i);
			p.addArc(pSource, tSAnd);
			p.addArc(tEAnd, pSink);

			startEndPlaces.add(pSource);
			startEndPlaces.add(pSink);
		} else if (node instanceof Operator && ((Operator) node).getName().equals("Xor")) {
			i.incrementAndGet();
			List<Place> node1 = construPetriNetRec(p, ((Operator) node).getChild1(), activityMap, activityXorMap, i);
			i.incrementAndGet();
			List<Place> node2 = construPetriNetRec(p, ((Operator) node).getChild2(), activityMap, activityXorMap, i);

			Place pSinkNode2 = node2.get(1);
			Place pSourceNode2 = node2.get(0);
			Place pSinkNode1 = node1.get(1);
			Place pSourceNode1 = node1.get(0);

			Transition tSXorN1 = p.addTransition("tau start 1n" + i);
			tSXorN1.setInvisible(true);
			p.addArc(tSXorN1, pSourceNode1);

			Transition tEXorN1 = p.addTransition("tau end 1n" + i);
			tEXorN1.setInvisible(true);
			p.addArc(pSinkNode1, tEXorN1);

			Transition tSXorN2 = p.addTransition("tau start 2n" + i);
			tSXorN2.setInvisible(true);
			p.addArc(tSXorN2, pSourceNode2);

			Transition tEXorN2 = p.addTransition("tau end 2n" + i);
			tEXorN2.setInvisible(true);
			p.addArc(pSinkNode2, tEXorN2);

			Place pSource = p.addPlace("source" + i);
			Place pSink = p.addPlace("sink" + i);
			p.addArc(pSource, tSXorN1);
			p.addArc(pSource, tSXorN2);
			p.addArc(tEXorN1, pSink);
			p.addArc(tEXorN2, pSink);

			startEndPlaces.add(pSource);
			startEndPlaces.add(pSink);
		} else if (node instanceof Operator && ((Operator) node).getName().equals("Loop")) {
			i.incrementAndGet();
			List<Place> node1 = construPetriNetRec(p, ((Operator) node).getChild1(), activityMap, activityXorMap, i);
			i.incrementAndGet();
			List<Place> node2 = construPetriNetRec(p, ((Operator) node).getChild2(), activityMap, activityXorMap, i);

			i.incrementAndGet();
			List<Place> node1End = construPetriNetRec(p, ((Operator) node).getChild1(), activityMap, activityXorMap, i);
			i.incrementAndGet();
			List<Place> node2End = construPetriNetRec(p, ((Operator) node).getChild2(), activityMap, activityXorMap, i);
			
			Place pSinkNode2 = node2.get(1);
			Place pSourceNode2 = node2.get(0);
			Place pSinkNode1 = node1.get(1);
			Place pSourceNode1 = node1.get(0);
			
			Place pSourceNode1E = node1End.get(0);
			Place pSinkNode1E = node1End.get(1);
			Place pSourceNode2E = node2End.get(0);
			Place pSinkNode2E = node2End.get(1);

			Place loopPlace = p.addPlace("loop" + i);

			Transition tSLoop = p.addTransition("tau start loop" + i);
			tSLoop.setInvisible(true);
			p.addArc(loopPlace, tSLoop);
			p.addArc(tSLoop, pSourceNode2);

			Transition tELoop = p.addTransition("tau end loop" + i);
			tELoop.setInvisible(true);
			p.addArc(tELoop, pSourceNode1);
			p.addArc(pSinkNode2, tELoop);

			Transition tMLoop = p.addTransition("tau mid loop" + i);
			tMLoop.setInvisible(true);
			p.addArc(pSinkNode1, tMLoop);
			p.addArc(tMLoop, loopPlace);

			Transition tSExt = p.addTransition("tau start ext" + i);
			tSExt.setInvisible(true);
			p.addArc(loopPlace, tSExt);
			p.addArc(tSExt, pSourceNode2E);
			
			Transition tEExt = p.addTransition("tau end ext" + i);
			tEExt.setInvisible(true);
			p.addArc(pSinkNode2E, tEExt);
			p.addArc(tEExt, pSourceNode1E);
			
			Place pSource = p.addPlace("source" + i);
			Place pSink = p.addPlace("sink" + i);

			Transition tSAll = p.addTransition("tau start all" + i);
			tSAll.setInvisible(true);
			p.addArc(pSource, tSAll);
			p.addArc(tSAll, pSourceNode1);

			Transition tEAll = p.addTransition("tau end all" + i);
			tEAll.setInvisible(true);
			p.addArc(pSinkNode1E, tEAll);
			p.addArc(tEAll, pSink);

			startEndPlaces.add(pSource);
			startEndPlaces.add(pSink);
		}

		return startEndPlaces;
	}

	static public AcceptingPetriNet linearFormatToAPN(PluginContext context, List<Object> linearFormat,
			Map<Integer, String> activityMap, Map<Integer, List<String>> activityXorMap) {
		Petrinet p = new PetrinetImpl("");
		Deque<Place> stack = new ArrayDeque<Place>();
		for (int i = 0; i < linearFormat.size(); i++) {
			Object current = linearFormat.get(i);
			if (current instanceof Leaf) {
				Leaf currentLeaf = (Leaf) current;
				Place pSource = p.addPlace("source" + i);
				Place pSink = p.addPlace("sink" + i);
				if (activityMap.containsKey(currentLeaf.getName())) {

					Transition t = p.addTransition(activityMap.get(currentLeaf.getName()));
					p.addArc(pSource, t);
					p.addArc(t, pSink);
				} else if (activityXorMap.containsKey(currentLeaf.getName())) {
					List<String> xorActivities = activityXorMap.get(currentLeaf.getName());
					Transition t1 = p.addTransition(xorActivities.get(0));
					Transition t2 = p.addTransition(xorActivities.get(1));
					p.addArc(pSource, t1);
					p.addArc(pSource, t2);
					p.addArc(t1, pSink);
					p.addArc(t2, pSink);
				}
				stack.offerFirst(pSource);
				stack.offerFirst(pSink);
			} else if (current instanceof Operator && ((Operator) current).getName().equals("Seq")) {
				Place pSinkNode2 = stack.pollFirst();
				Place pSourceNode2 = stack.pollFirst();
				Place pSinkNode1 = stack.pollFirst();
				Place pSourceNode1 = stack.pollFirst();
				Transition tMSeq = p.addTransition("tau seq" + i);
				tMSeq.setInvisible(true);
				p.addArc(pSinkNode1, tMSeq);
				p.addArc(tMSeq, pSourceNode2);

				Transition tSSeq = p.addTransition("tau start seq" + i);
				tSSeq.setInvisible(true);
				p.addArc(tSSeq, pSourceNode1);

				Transition tESeq = p.addTransition("tau end seq" + i);
				tESeq.setInvisible(true);
				p.addArc(pSinkNode2, tESeq);

				Place pSource = p.addPlace("source" + i);
				Place pSink = p.addPlace("sink" + i);
				p.addArc(pSource, tSSeq);
				p.addArc(tESeq, pSink);
				stack.offerFirst(pSource);
				stack.offerFirst(pSink);
			} else if (current instanceof Operator && ((Operator) current).getName().equals("And")) {
				Place pSinkNode2 = stack.pollFirst();
				Place pSourceNode2 = stack.pollFirst();
				Place pSinkNode1 = stack.pollFirst();
				Place pSourceNode1 = stack.pollFirst();

				Transition tSAnd = p.addTransition("tau start and" + i);
				tSAnd.setInvisible(true);
				p.addArc(tSAnd, pSourceNode1);
				p.addArc(tSAnd, pSourceNode2);

				Transition tEAnd = p.addTransition("tau end and" + i);
				tEAnd.setInvisible(true);
				p.addArc(pSinkNode1, tEAnd);
				p.addArc(pSinkNode2, tEAnd);
				Place pSource = p.addPlace("source" + i);
				Place pSink = p.addPlace("sink" + i);
				p.addArc(pSource, tSAnd);
				p.addArc(tEAnd, pSink);

				stack.offerFirst(pSource);
				stack.offerFirst(pSink);
			} else if (current instanceof Operator && ((Operator) current).getName().equals("Xor")) {
				Place pSinkNode2 = stack.pollFirst();
				Place pSourceNode2 = stack.pollFirst();
				Place pSinkNode1 = stack.pollFirst();
				Place pSourceNode1 = stack.pollFirst();

				Transition tSXorN1 = p.addTransition("tau start 1n" + i);
				tSXorN1.setInvisible(true);
				p.addArc(tSXorN1, pSourceNode1);

				Transition tEXorN1 = p.addTransition("tau end 1n" + i);
				tEXorN1.setInvisible(true);
				p.addArc(pSinkNode1, tEXorN1);

				Transition tSXorN2 = p.addTransition("tau start 2n" + i);
				tSXorN2.setInvisible(true);
				p.addArc(tSXorN2, pSourceNode2);

				Transition tEXorN2 = p.addTransition("tau end 2n" + i);
				tEXorN2.setInvisible(true);
				p.addArc(pSinkNode2, tEXorN2);

				Place pSource = p.addPlace("source" + i);
				Place pSink = p.addPlace("sink" + i);
				p.addArc(pSource, tSXorN1);
				p.addArc(pSource, tSXorN2);
				p.addArc(tEXorN1, pSink);
				p.addArc(tEXorN2, pSink);

				stack.offerFirst(pSource);
				stack.offerFirst(pSink);
			} else if (current instanceof Operator && ((Operator) current).getName().equals("Loop")) {
				Place pSinkNode2 = stack.pollFirst();
				Place pSourceNode2 = stack.pollFirst();
				Place pSinkNode1 = stack.pollFirst();
				Place pSourceNode1 = stack.pollFirst();

				Place loopPlace = p.addPlace("loop" + i);

				Transition tSLoop = p.addTransition("tau start loop" + i);
				tSLoop.setInvisible(true);
				p.addArc(loopPlace, tSLoop);
				p.addArc(tSLoop, pSourceNode2);

				Transition tELoop = p.addTransition("tau end loop" + i);
				tELoop.setInvisible(true);
				p.addArc(tELoop, pSourceNode1);
				p.addArc(pSinkNode2, tELoop);

				Transition tMLoop = p.addTransition("tau mid loop" + i);
				tMLoop.setInvisible(true);
				p.addArc(pSinkNode1, tMLoop);
				p.addArc(tMLoop, loopPlace);

				Place pSource = p.addPlace("source" + i);
				Place pSink = p.addPlace("sink" + i);

				Transition tSAll = p.addTransition("tau start all" + i);
				tSAll.setInvisible(true);
				p.addArc(pSource, tSAll);
				p.addArc(tSAll, pSourceNode1);

				Transition tEAll = p.addTransition("tau end all" + i);
				tEAll.setInvisible(true);
				p.addArc(loopPlace, tEAll);
				p.addArc(tEAll, pSink);

				stack.offerFirst(pSource);
				stack.offerFirst(pSink);
			}
		}

		Place pSink = stack.pollFirst();
		Place pSource = stack.pollFirst();
		Marking initialMarking = new Marking();
		initialMarking.add(pSource);
		Marking finalMarking = new Marking();
		finalMarking.add(pSink);
		AcceptingPetriNet ap = new AcceptingPetriNetImpl(p, initialMarking, finalMarking);

		ReduceUsingMurataRulesPlugin murata = new ReduceUsingMurataRulesPlugin();
		ReduceUsingMurataRulesParameters murataParams = new ReduceUsingMurataRulesParameters();
		murataParams.setRetainBehavior(true);
		murataParams.setTryConnections(false);
		ap = murata.run(context, ap, murataParams);
		return ap;
	}

	public static List<Object> copyLinearFormat(Node root) {
		List<Object> copyLinearFormat = new ArrayList<Object>();
		construLinearFormatRec(copyLinearFormat, null, root, false);
		return copyLinearFormat;
	}

	private static void construLinearFormatRec(List<Object> copyLinearFormat, Operator parent, Node node,
			boolean leftChild) {
		if (node instanceof Leaf) {
			Leaf n = (Leaf) node;
			Leaf l = new Leaf(n.getName());
			l.setDepth(n.getDepth());
			l.setParent(parent);
			l.setRepresLeaf(l.getName());
			if (parent != null) {
				if (leftChild)
					parent.setChild1(l);
				else
					parent.setChild2(l);
			}
			copyLinearFormat.add(l);
		} else if (node instanceof Operator) {
			Operator n = (Operator) node;
			Operator o = new Operator(n.getName());
			if (parent != null) {
				if (leftChild)
					parent.setChild1(o);
				else
					parent.setChild2(o);
			}
			o.setRepresLeaf(n.getRepresLeaf());
			construLinearFormatRec(copyLinearFormat, o, n.getChild1(), true);
			construLinearFormatRec(copyLinearFormat, o, n.getChild2(), false);
			copyLinearFormat.add(o);
		}
	}
}